@@ -788,6 +788,7 @@ def match_files_except_lines(file1, file2, excluded=None):
 class RerunLock:
     """Lock process from reentering when seeing lock file on disk."""
     def __init__(self, name, folder=None, infohook=_logger.info, warnhook=_logger.warning, errorhook=_logger.error):
+        os.makedirs(folder, exist_ok=True)
         filename = 'lock_{}'.format(name) if name else 'lock_{}'.format(next(tempfile._get_candidate_names()))
         self.lockFile = join(folder, filename) if folder else join(_script_dir, os.pardir, filename)
         self.infoHook = infohook
@@ -919,8 +920,188 @@ def remove_from_os_paths(bindir):
     os.environ['PATH'] = os.environ['PATH'].replace(bindir, '')
 
 
+def run_cmd(cmd, cwd='.'):
+    try:
+        proc = subprocess.run(cmd, check=True, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
+        _logger.debug(proc.stdout.decode(TXT_CODEC))
+    except subprocess.CalledProcessError as e:
+        _logger.error(f'stderr: {e.stderr.decode(TXT_CODEC)}')
+        raise e
+
+
+def run_daemon(cmd, cwd='.'):
+    proc = None
+    try:
+        proc = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
+    except subprocess.CalledProcessError as e:
+        _logger.error(f'stderr: {e.stderr.decode(TXT_CODEC)}')
+        raise e
+    return proc
+
+
+
+def extract_call_args(file, caller, callee):
+    """
+    - only support literal args
+    - will throw if an arg value is a function call itself
+    """
+    import ast
+    import importlib
+    import inspect
+    mod = splitext(basename(file))[0]
+    sys.path.insert(0, dirname(file))
+    mod = importlib.import_module(mod)
+    parsed = ast.parse(inspect.getsource(mod))
+    raw_calls = { # lineno, args, keywords
+        'func': [],
+        'method': []
+    }
+    for node in parsed.body:
+        if not isinstance(node, ast.FunctionDef) \
+            or node.name != caller:
+            continue
+        # hit function definition: add_arguments
+        # next: find function calls in this function
+        caller_def = node
+        callee_calls = []
+        for node in caller_def.body:
+            if not isinstance(node.value, ast.Call):
+                continue
+            print(node.value.func)
+            # hit a function/method call
+            call_type = None
+            found_func_call = isinstance(node.value.func, ast.Name) and node.value.func.id == callee
+            found_method_call = isinstance(node.value.func, ast.Attribute) and node.value.func.attr == callee
+            if found_func_call:
+                call_type = 'func'
+            elif found_method_call:
+                call_type = 'method'
+            else:
+                continue
+            raw_calls[call_type].append(node.value)
+
+    # collect args and kwargs
+    calls = { # lineno, args, keywords
+        'func': [],
+        'method': []
+    }
+    def get_kwarg_value_const(kw):
+        return kw.value.value
+    def get_kwarg_value_name(kw):
+        return kw.value.id
+    def get_kwarg_value_list(kw):
+        nodes = kw.value.elts
+        values = []
+        for node in nodes:
+            if not isinstance(node, ast.Constant):
+                values.append(None)
+                continue
+            values.append(node.value)
+        return values
+    def get_kwarg_value_misc(kw):
+        return None
+
+    for calltype, rc in raw_calls.items():
+        for call in rc:
+            args = []
+            kwargs = []
+            for arg in call.args:
+                args.append(arg.value)
+            for kw in call.keywords:
+                key = kw.arg
+                if isinstance(kw.value, ast.Constant):
+                    value = get_kwarg_value_const(kw)
+                elif isinstance(kw.value, ast.Name):
+                    value = get_kwarg_value_name(kw)
+                elif isinstance(kw.value, (ast.List, ast.Tuple)):
+                    value = get_kwarg_value_list(kw)
+                else:
+                    print(f'Unsupported syntax node: {kw.value}. Will fallback to None.')
+                    value = None
+                kwargs.append((key, value))
+            call = {
+                'args': args,
+                'kwargs': {k: v for k, v in kwargs}
+            }
+            calls[calltype].append(call)
+    sys.modules.pop(mod)
+    return calls['func'], calls['method']
+
+
+def extract_class_attributes(file, classname):
+    """
+    assume
+    - class is defined at the top-level of source file
+    - attributes all use type-annotated assignemts
+    - all assgiments are from attributes
+    """
+    import ast
+    import importlib
+    import inspect
+    mod = splitext(basename(file))[0]
+    sys.path.insert(0, dirname(file))
+    mod = importlib.import_module(mod)
+    parsed = ast.parse(inspect.getsource(mod))
+    
+    class_node = None
+    for node in parsed.body:
+        if not isinstance(node, ast.ClassDef):
+            continue
+        if node.name != classname:
+            continue
+        class_node = node
+    if not class_node:
+        return None
+    # found class and parse
+    names = [node.attr for node in ast.walk(class_node) if isinstance(node, ast.Attribute)]
+    types = [node.annotation.id for node in ast.walk(class_node) if isinstance(node, ast.AnnAssign)]
+    values = []
+    for node in ast.walk(class_node):
+        if not isinstance(node, ast.AnnAssign):
+            continue
+        if isinstance(node.value, ast.Constant):
+            values.append(node.value.value)
+        elif isinstance(node.value, (ast.List, ast.Tuple)):
+            nodes = node.value.elts
+            raw_values = []
+            for node in nodes:
+                # only support constants inside list
+                # non-consts are taken as None
+                rv = node.value if isinstance(node, ast.Constant) else None
+                raw_values.append(rv)
+            values.append(raw_values)
+    attributes = [{'name': n, 'type': t, 'value': v} for n, t, v in zip(names, types, values)]
+    sys.modules.pop(mod)
+    return attributes
+
+
+def substitute_lines_between_keywords(lines, file, opkey, edkey):
+    """
+    assume input lines all have line ends
+    """
+    all_lines = []
+    with open(file) as fp:
+        all_lines = fp.readlines()
+    # find range
+    rg_insert = [-1, -1]
+    for l, line in enumerate(all_lines):
+        raw = line.strip()
+        if raw.startswith(opkey):
+            rg_insert[0] = l
+        if raw.startswith(edkey):
+            rg_insert[1] = l
+            break
+    # remove lines in b/w
+    n_lines_btw = rg_insert[1] - rg_insert[0]
+    if n_lines_btw > 1:
+        del all_lines[rg_insert[0]+1 : rg_insert[1]]
+    all_lines[rg_insert[0]+1 : rg_insert[0]+1] = lines
+    with open(file, 'w') as fp:
+        fp.writelines(all_lines)
+
+
 def _test():
-    pass
+    print(extract_class_attributes('/Users/bin.luo/Desktop/_dev/miatech/create_py_proj/test/proto_gen/_ref_org/core.py', 'Output'))
 
 
 if __name__ == '__main__':
